@@ -2,11 +2,10 @@
 /**
  * Turba directory driver implementation for PHP's LDAP extension.
  *
- * $Horde: turba/lib/Driver/ldap.php,v 1.54.4.9 2005/06/23 21:31:11 selsky Exp $
+ * $Horde: turba/lib/Driver/ldap.php,v 1.54.4.15 2006/07/27 01:48:28 chuck Exp $
  *
  * @author  Chuck Hagenbuch <chuck@horde.org>
  * @author  Jon Parise <jon@csh.rit.edu>
- * @since   Turba 0.0.1
  * @package Turba
  */
 class Turba_Driver_ldap extends Turba_Driver {
@@ -18,13 +17,6 @@ class Turba_Driver_ldap extends Turba_Driver {
      */
     var $_ds = 0;
 
-    /**
-     * Net_LDAP schema object.
-     *
-     * @var Net_LDAP_Schema
-     */
-    var $_schema;
-
     /**
      * Constructs a new Turba LDAP driver object.
      *
@@ -188,7 +180,7 @@ class Turba_Driver_ldap extends Turba_Driver {
         if (is_array($dn)) {
             $results = array();
             foreach ($dn as $d) {
-                $res = @ldap_read($this->_ds, $d, $filter, $attr);
+                $res = @ldap_read($this->_ds, String::convertCharset($d, NLS::getCharset(), $this->_params['charset']), $filter, $attr);
                 if ($res) {
                     if (!is_a($result = $this->_getResults($fields, $res), 'PEAR_Error')) {
                         $results = array_merge($results, $result);
@@ -202,7 +194,7 @@ class Turba_Driver_ldap extends Turba_Driver {
             return $results;
         }
 
-        $res = @ldap_read($this->_ds, $dn, $filter, $attr);
+        $res = @ldap_read($this->_ds, String::convertCharset($dn, NLS::getCharset(), $this->_params['charset']), $filter, $attr);
         if (!$res) {
             return PEAR::raiseError(sprintf(_("Read failed: (%s) %s"), ldap_errno($this->_ds), ldap_error($this->_ds)));
         }
@@ -256,14 +248,9 @@ class Turba_Driver_ldap extends Turba_Driver {
             }
         }
 
-        /* Encode entries. */
-        foreach ($attributes as $key => $val) {
-            if (!is_array($val)) {
-                $attributes[$key] = String::convertCharset($val, NLS::getCharset(), $this->_params['charset']);
-            }
-        }
+        $this->_encodeAttributes($attributes);
 
-        if (!@ldap_add($this->_ds, $dn, $attributes)) {
+        if (!@ldap_add($this->_ds, String::convertCharset($dn, NLS::getCharset(), $this->_params['charset']), $attributes)) {
             return PEAR::raiseError('Failed to add an object: [' . ldap_errno($this->_ds) . '] "' . ldap_error($this->_ds) . '" DN: ' . $dn . ' (attributes: [' . serialize($attributes) . ']).' . "Charset:" . NLS::getCharset());
         } else {
             return true;
@@ -279,7 +266,7 @@ class Turba_Driver_ldap extends Turba_Driver {
             return PEAR::raiseError(_("Invalid key specified."));
         }
 
-        if (!@ldap_delete($this->_ds, $object_id)) {
+        if (!@ldap_delete($this->_ds, String::convertCharset($object_id, NLS::getCharset(), $this->_params['charset']))) {
             return PEAR::raiseError(sprintf(_("Delete failed: (%s) %s"), ldap_errno($this->_ds), ldap_error($this->_ds)));
         } else {
             return true;
@@ -297,25 +284,19 @@ class Turba_Driver_ldap extends Turba_Driver {
          * values. These are needed so that we can delete any
          * attributes that have been removed by using ldap_mod_del. */
         $filter = $this->_buildObjectclassFilter();
-        $oldres = @ldap_read($this->_ds, $object_id, $filter, array_keys($attributes));
+        $oldres = @ldap_read($this->_ds, String::convertCharset($object_id, NLS::getCharset(), $this->_params['charset']), $filter, array_merge(array_keys($attributes), array('objectclass')));
         $info = ldap_get_attributes($this->_ds, ldap_first_entry($this->_ds, $oldres));
 
         /* Check if we need to rename the object. */
         if ($this->_params['version'] == 3 &&
             String::lower($this->_makeKey($attributes)) != String::lower($object_id)) {
-            if (isset($this->_params['dn']) && is_array($this->_params['dn']) && count($this->_params['dn'])) {
-                $pairs = array();
-                foreach ($this->_params['dn'] as $param) {
-                    if (isset($attributes[$param])) {
-                        $pairs[] = array($param, $attributes[$param]);
-                    }
-                }
-                $newrdn = $this->_quoteDN($pairs);
-            } else {
+            $newrdn = $this->_makeRDN($attributes);
+            if ($newrdn == '') {
                 return PEAR::raiseError(_("Missing DN in LDAP source configuration."));
             }
 
-            if (ldap_rename($this->_ds, $object_id, $newrdn, $this->_params['root'], true)) {
+            if (ldap_rename($this->_ds, String::convertCharset($object_id, NLS::getCharset(), $this->_params['charset']),
+                    String::convertCharset($newrdn, NLS::getCharset(), $this->_params['charset']), $this->_params['root'], true)) {
                 $object_id = $newrdn . ',' . $this->_params['root'];
             } else {
                 return PEAR::raiseError(sprintf(_("Failed to change name: (%s) %s; Old DN = %s, New DN = %s, Root = %s"),
@@ -323,10 +304,9 @@ class Turba_Driver_ldap extends Turba_Driver {
             }
         }
 
-        /* The attributes in the attributes array are all lower case
-         * while they are mixedCase in the search result. So convert
-         * the keys to lower. */
+        /* Work only with lowercase keys. */
         $info = array_change_key_case($info, CASE_LOWER);
+        $attributes = array_change_key_case($attributes, CASE_LOWER);
 
         foreach ($info as $key => $value) {
             $var = $info[$key];
@@ -340,30 +320,53 @@ class Turba_Driver_ldap extends Turba_Driver {
                 $attributes[$key] == '') {
 
                 $oldval[$key] = $var[0];
-                if (!@ldap_mod_del($this->_ds, $object_id, $oldval)) {
+                if (!@ldap_mod_del($this->_ds, String::convertCharset($object_id, NLS::getCharset(), $this->_params['charset']), $oldval)) {
                     return PEAR::raiseError(sprintf(_("Modify failed: (%s) %s"), ldap_errno($this->_ds), ldap_error($this->_ds)));
                 }
                 unset($attributes[$key]);
             }
         }
 
-        /* Encode entries. */
-        unset($attributes[$object_key]);
-        foreach ($attributes as $key => $val) {
-            if (!is_array($val)) {
-                $attributes[$key] = String::convertCharset($val, NLS::getCharset(), $this->_params['charset']);
-            }
-        }
-
+        unset($attributes[String::lower($object_key)]);
+        $this->_encodeAttributes($attributes);
         $attributes = array_filter($attributes, array($this, '_emptyAttributeFilter'));
 
-        if (!@ldap_modify($this->_ds, $object_id, $attributes)) {
+        /* Modify objectclass if old one is outdated. */
+        $attributes['objectclass'] = array_unique(array_map('strtolower', array_merge($info['objectclass'], $this->_params['objectclass'])));
+        unset($attributes['objectclass']['count']);
+        $attributes['objectclass'] = array_values($attributes['objectclass']);
+
+        if (!@ldap_modify($this->_ds, String::convertCharset($object_id, NLS::getCharset(), $this->_params['charset']), $attributes)) {
             return PEAR::raiseError(sprintf(_("Modify failed: (%s) %s"), ldap_errno($this->_ds), ldap_error($this->_ds)));
         } else {
             return $object_id;
         }
     }
 
+    /**
+     * Build a RDN based on a set of attributes and what attributes
+     * make a RDN for the current source.
+     *
+     * @param array $attributes The attributes (in driver keys) of the
+     *                          object being added.
+     *
+     * @return string  The RDN for the new object.
+     */
+    function _makeRDN($attributes)
+    {
+        if (!is_array($this->_params['dn'])) {
+            return '';
+        }
+
+        $pairs = array();
+        foreach ($this->_params['dn'] as $param) {
+            if (isset($attributes[$param])) {
+                $pairs[] = array($param, $attributes[$param]);
+            }
+        }
+        return $this->_quoteDN($pairs);
+    }
+
     /**
      * Build a DN based on a set of attributes and what attributes
      * make a DN for the current source.
@@ -375,18 +378,7 @@ class Turba_Driver_ldap extends Turba_Driver {
      */
     function _makeKey($attributes)
     {
-        $dn = '';
-        if (is_array($this->_params['dn'])) {
-            $pairs = array();
-            foreach ($this->_params['dn'] as $param) {
-                if (isset($attributes[$param])) {
-                    $pairs[] = array($param, $attributes[$param]);
-                }
-            }
-            $dn = $this->_quoteDN($pairs);
-        }
-
-        return $dn . ',' . $this->_params['root'];
+        return $this->_makeRDN($attributes) . ',' . $this->_params['root'];
     }
 
     /**
@@ -451,7 +443,7 @@ class Turba_Driver_ldap extends Turba_Driver {
             foreach ($fields as $field) {
                 $field_l = String::lower($field);
                 if ($field == 'dn') {
-                    $result[$field] = $entry[$field_l];
+                    $result[$field] = String::convertCharset($entry[$field_l], $this->_params['charset']);
                 } else {
                     $result[$field] = '';
                     if (!empty($entry[$field_l])) {
@@ -461,6 +453,21 @@ class Turba_Driver_ldap extends Turba_Driver {
                             }
                             $result[$field] .= String::convertCharset($entry[$field_l][$j], $this->_params['charset']);
                         }
+
+                        /* If schema checking is enabled check the
+                         * backend syntax. */
+                        if (!empty($this->_params['checksyntax'])) {
+                            $postal = $this->_isPostalAddress($field_l);
+                        } else {
+                            /* Otherwise rely on the attribute mapping
+                             * in attributes.php. */
+                            $attr = array_search($field_l, $this->map);
+                            $postal = (!empty($attr) && !empty($GLOBALS['attributes'][$attr]) &&
+                                       $GLOBALS['attributes'][$attr]['type'] == 'address');
+                        }
+                        if ($postal) {
+                            $result[$field] = str_replace('$', "\r\n", $result[$field]);
+                        }
                     }
                 }
             }
@@ -492,6 +499,34 @@ class Turba_Driver_ldap extends Turba_Driver {
         }
     }
 
+    /**
+     * Format and encode attributes including postal addresses,
+     * character set encoding, etc.
+     */
+    function _encodeAttributes(&$attributes)
+    {
+        foreach ($attributes as $key => $val) {
+            /* If schema checking is enabled check the backend syntax. */
+            if (!empty($this->_params['checksyntax'])) {
+                $postal = $this->_isPostalAddress($key);
+            } else {
+                /* Otherwise rely on the attribute mapping in
+                 * attributes.php. */
+                $attr = array_search($key, $this->map);
+                $postal = (!empty($attr) && !empty($val) && !empty($GLOBALS['attributes'][$attr]) &&
+                           $GLOBALS['attributes'][$attr]['type'] == 'address');
+            }
+            if ($postal) {
+                /* Correctly store postal addresses. */
+                $val = str_replace(array("\r\n", "\r", "\n"), '$', $val);
+            }
+
+            if (!is_array($val)) {
+                $attributes[$key] = String::convertCharset($val, NLS::getCharset(), $this->_params['charset']);
+            }
+        }
+    }
+
     /**
      * Build an LDAP filter based on the objectclass parameter.
      *
@@ -527,30 +562,19 @@ class Turba_Driver_ldap extends Turba_Driver {
      */
     function _checkRequiredAttributes($objectclasses)
     {
-       require_once 'Net/LDAP.php';
-
-       $config = array('host' => $this->_params['server'],
-                       'port' => $this->_params['port']);
+       $schema = &$this->_getSchema();
+       if (is_a($schema, 'PEAR_Error')) {
+           return $schema;
+       }
 
        $retval = array();
 
-       $ldap = &new Net_LDAP($config);
-       $ldap->_link = $this->_ds;
-
-       if (!is_object($this->_schema)) {
-           $this->_schema = &$ldap->schema();
-       }
-
-       if (is_a($this->_schema, 'PEAR_Error')) {
-           return $this->_schema;
-       }
-
        foreach ($objectclasses as $oc) {
            if (String::lower($oc) == 'top') {
                continue;
            }
 
-           $required = $this->_schema->must($oc);
+           $required = $schema->must($oc);
 
            if (is_array($required)) {
                foreach ($required as $v) {
@@ -589,13 +613,30 @@ class Turba_Driver_ldap extends Turba_Driver {
                           15 => 1, /* Directory string. */
                           );
 
-        if (preg_match('/(\d+)\{\d+\}$/', $syntax, $matches) &&
-            isset($okSyntax[$matches[1]])) {
+        if (preg_match('/^(.*)\.(\d+)\{\d+\}$/', $syntax, $matches) &&
+            $matches[1] == "1.3.6.1.4.1.1466.115.121.1" &&
+            isset($okSyntax[$matches[2]])) {
             return true;
         }
         return false;
     }
 
+    /**
+     * Checks if an attribute refers to a Postal Address.
+     *
+     * @access private
+     *
+     * @param string $attribute  An attribute name.
+     *
+     * @return boolean  True if the specified attribute refers to a Postal Address.
+     */
+    function _isPostalAddress($attribute)
+    {
+        /* LDAP postal address syntax is
+         * 1.3.6.1.4.1.1466.115.121.1.41 */
+        return $this->_getSyntax($attribute) == '1.3.6.1.4.1.1466.115.121.1.41';
+    }
+
     /**
      * Returns the syntax of an attribute, if necessary recursively.
      *
@@ -607,8 +648,12 @@ class Turba_Driver_ldap extends Turba_Driver {
      */
     function _getSyntax($att)
     {
-        $attv = $this->_schema->get('attribute', $att);
+        $schema = &$this->_getSchema();
+        if (is_a($schema, 'PEAR_Error')) {
+            return $schema;
+        }
 
+        $attv = $schema->get('attribute', $att);
         if (isset($attv['syntax'])) {
             return $attv['syntax'];
         } else {
@@ -616,6 +661,34 @@ class Turba_Driver_ldap extends Turba_Driver {
         }
     }
 
+    /**
+     * Returns an LDAP_Schema object that containts the LDAP schema.
+     *
+     * @access private
+     *
+     * @return Net_LDAP_Schema  Returns a reference to the ldap schema object.
+     */
+    function &_getSchema()
+    {
+        require_once 'Net/LDAP.php';
+
+        static $_schema;
+
+        /* Check if the cached schema is valid, */
+        if (is_object($_schema) && is_a($_schema, 'Net_LDAP_Schema')) {
+            return $_schema;
+        }
+
+        $config = array('host' => $this->_params['server'],
+                        'port' => $this->_params['port']);
+        $ldap = &new Net_LDAP($config);
+        $ldap->_link = $this->_ds;
+
+        $_schema = &$ldap->schema();
+
+        return $_schema;
+    }
+
     /**
      * Take an array of DN elements and properly quote it according to
      * RFC 1485.
